import os
import pytest
import fiona
from fiona.model import Feature

from .conftest import requires_gpkg

example_schema = {
    "geometry": "Point",
    "properties": [("title", "str")],
}

example_crs = {
    "a": 6370997,
    "lon_0": -100,
    "y_0": 0,
    "no_defs": True,
    "proj": "laea",
    "x_0": 0,
    "units": "m",
    "b": 6370997,
    "lat_0": 45,
}

example_features = [
    Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
            "properties": {"title": "One"},
        }
    ),
    Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": [1.0, 2.0]},
            "properties": {"title": "Two"},
        }
    ),
    Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": [3.0, 4.0]},
            "properties": {"title": "Three"},
        }
    ),
]


@requires_gpkg
def test_read_gpkg(path_coutwildrnp_gpkg):
    """
    Implicitly tests writing gpkg as the fixture will create the data source on
    first request
    """
    with fiona.open(path_coutwildrnp_gpkg, "r") as src:
        assert len(src) == 67
        feature = next(iter(src))
        assert feature.geometry["type"] == "Polygon"
        assert feature.properties["NAME"] == "Mount Naomi Wilderness"


@requires_gpkg
def test_write_gpkg(tmpdir):
    path = str(tmpdir.join("foo.gpkg"))

    with fiona.open(
        path, "w", driver="GPKG", schema=example_schema, crs=example_crs
    ) as dst:
        dst.writerecords(example_features)

    with fiona.open(path) as src:
        assert src.schema["geometry"] == "Point"
        assert len(src) == 3


@requires_gpkg
def test_write_multilayer_gpkg(tmpdir):
    """
    Test that writing a second layer to an existing geopackage doesn't remove
    and existing layer for the dataset.
    """
    path = str(tmpdir.join("foo.gpkg"))

    with fiona.open(
        path, "w", driver="GPKG", schema=example_schema, layer="layer1", crs=example_crs
    ) as dst:
        dst.writerecords(example_features[0:2])

    with fiona.open(
        path, "w", driver="GPKG", schema=example_schema, layer="layer2", crs=example_crs
    ) as dst:
        dst.writerecords(example_features[2:])

    with fiona.open(path, layer="layer1") as src:
        assert src.schema["geometry"] == "Point"
        assert len(src) == 2

    with fiona.open(path, layer="layer2") as src:
        assert src.schema["geometry"] == "Point"
        assert len(src) == 1



"""Tests for `$ fio calc`."""


import json

from click.testing import CliRunner

from fiona.fio.main import main_group


def test_fail():
    runner = CliRunner()
    result = runner.invoke(main_group, ['calc', "TEST", "f.properties.test > 5"],
                           '{"type": "no_properties"}')
    assert result.exit_code == 1


def _load(output):
    features = []
    for x in output.splitlines():
        try:
            features.append(json.loads(x))
        except:
            # Click combines stdout and stderr and shapely dumps logs to
            # stderr that are not JSON
            # https://github.com/pallets/click/issues/371
            pass
    return features


def test_calc_seq(feature_seq, runner):
    result = runner.invoke(main_group, ['calc', 
        "TEST",
        "f.properties.AREA / f.properties.PERIMETER"],
        feature_seq)
    assert result.exit_code == 0

    feats = _load(result.output)
    assert len(feats) == 2
    for feat in feats:
        assert feat['properties']['TEST'] == \
            feat['properties']['AREA'] / feat['properties']['PERIMETER']


def test_bool_seq(feature_seq, runner):
    result = runner.invoke(main_group, ['calc', "TEST", "f.properties.AREA > 0.015"],
                           feature_seq)
    assert result.exit_code == 0
    feats = _load(result.output)
    assert len(feats) == 2
    assert feats[0]['properties']['TEST']
    assert not feats[1]['properties']['TEST']


def test_existing_property(feature_seq, runner):
    result = runner.invoke(
        main_group, ["calc", "AREA", "f.properties.AREA * 2"], feature_seq
    )
    assert result.exit_code == 2

    result = runner.invoke(main_group, ['calc', "--overwrite", "AREA", "f.properties.AREA * 2"],
                           feature_seq)
    assert result.exit_code == 0
    feats = _load(result.output)
    assert len(feats) == 2
    for feat in feats:
        assert 'AREA' in feat['properties']



"""Tests of MemoryFile and ZippedMemoryFile"""

import os
from collections import OrderedDict
from io import BytesIO
import os

import pytest
import fiona
from fiona import supported_drivers
from fiona.drvsupport import _driver_supports_mode
from fiona.errors import DriverError
from fiona.io import MemoryFile, ZipMemoryFile
from fiona.meta import supports_vsi

from .conftest import requires_gdal2, requires_gpkg


@pytest.fixture(scope='session')
def profile_first_coutwildrnp_shp(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as col:
        return col.profile, next(iter(col))


@pytest.fixture(scope='session')
def data_coutwildrnp_json(path_coutwildrnp_json):
    with open(path_coutwildrnp_json, 'rb') as f:
        return f.read()


def test_memoryfile_ext():
    """File extensions are handled"""
    assert MemoryFile(ext=".geojson").name.endswith(".geojson")


def test_memoryfile_bare_ext():
    """File extensions without a leading . are handled"""
    assert MemoryFile(ext="geojson").name.endswith(".geojson")


def test_memoryfile_init(data_coutwildrnp_json):
    """In-memory GeoJSON file can be read"""
    with MemoryFile(data_coutwildrnp_json) as memfile:
        with memfile.open() as collection:
            assert len(collection) == 67


def test_memoryfile_incr_init(data_coutwildrnp_json):
    """In-memory GeoJSON file written in 2 parts can be read"""
    with MemoryFile() as memfile:
        memfile.write(data_coutwildrnp_json[:1000])
        memfile.write(data_coutwildrnp_json[1000:])
        with memfile.open() as collection:
            assert len(collection) == 67


def test_zip_memoryfile(bytes_coutwildrnp_zip):
    """In-memory zipped Shapefile can be read"""
    with ZipMemoryFile(bytes_coutwildrnp_zip) as memfile:
        with memfile.open('coutwildrnp.shp') as collection:
            assert len(collection) == 67


def test_zip_memoryfile_infer_layer_name(bytes_coutwildrnp_zip):
    """In-memory zipped Shapefile can be read with the default layer"""
    with ZipMemoryFile(bytes_coutwildrnp_zip) as memfile:
        with memfile.open() as collection:
            assert len(collection) == 67


def test_open_closed():
    """Get an exception when opening a dataset on a closed MemoryFile"""
    memfile = MemoryFile()
    memfile.close()
    assert memfile.closed
    with pytest.raises(OSError):
        memfile.open()


def test_open_closed_zip():
    """Get an exception when opening a dataset on a closed ZipMemoryFile"""
    memfile = ZipMemoryFile()
    memfile.close()
    assert memfile.closed
    with pytest.raises(OSError):
        memfile.open()


def test_write_memoryfile(profile_first_coutwildrnp_shp):
    """In-memory GeoJSON can be written"""
    profile, first = profile_first_coutwildrnp_shp
    profile['driver'] = 'GeoJSON'
    with MemoryFile() as memfile:
        with memfile.open(**profile) as col:
            col.write(first)
        memfile.seek(0)
        data = memfile.read()

    with MemoryFile(data) as memfile:
        with memfile.open() as col:
            assert len(col) == 1


@requires_gdal2
def test_memoryfile_write_extension(profile_first_coutwildrnp_shp):
    """In-memory shapefile gets an .shp extension by default"""
    profile, first = profile_first_coutwildrnp_shp
    profile['driver'] = 'ESRI Shapefile'
    with MemoryFile() as memfile:
        with memfile.open(**profile) as col:
            col.write(first)
        assert memfile.name.endswith(".shp")


def test_memoryfile_open_file_or_bytes_read(path_coutwildrnp_json):
    """Test MemoryFile.open when file_or_bytes has a read attribute """
    with open(path_coutwildrnp_json, 'rb') as f:
        with MemoryFile(f) as memfile:
            with memfile.open() as collection:
                assert len(collection) == 67


def test_memoryfile_bytesio(data_coutwildrnp_json):
    """GeoJSON file stored in BytesIO can be read"""
    with fiona.open(BytesIO(data_coutwildrnp_json)) as collection:
        assert len(collection) == 67


def test_memoryfile_fileobj(path_coutwildrnp_json):
    """GeoJSON file in an open file object can be read"""
    with open(path_coutwildrnp_json, 'rb') as f:
        with fiona.open(f) as collection:
            assert len(collection) == 67


def test_memoryfile_len(data_coutwildrnp_json):
    """Test MemoryFile.__len__ """
    with MemoryFile() as memfile:
        assert len(memfile) == 0
        memfile.write(data_coutwildrnp_json)
        assert len(memfile) == len(data_coutwildrnp_json)


def test_memoryfile_tell(data_coutwildrnp_json):
    """Test MemoryFile.tell() """
    with MemoryFile() as memfile:
        assert memfile.tell() == 0
        memfile.write(data_coutwildrnp_json)
        assert memfile.tell() == len(data_coutwildrnp_json)


def test_write_bytesio(profile_first_coutwildrnp_shp):
    """GeoJSON can be written to BytesIO"""
    profile, first = profile_first_coutwildrnp_shp
    profile['driver'] = 'GeoJSON'
    with BytesIO() as fout:
        with fiona.open(fout, 'w', **profile) as col:
            col.write(first)
        fout.seek(0)
        data = fout.read()

    with MemoryFile(data) as memfile:
        with memfile.open() as col:
            assert len(col) == 1


@requires_gpkg
def test_read_multilayer_memoryfile(path_coutwildrnp_json, tmpdir):
    """Test read access to multilayer dataset in from file-like object"""
    with fiona.open(path_coutwildrnp_json, "r") as src:
        schema = src.schema
        features = list(src)

    path = os.path.join(tmpdir, "test.gpkg")
    with fiona.open(path, "w", driver="GPKG", schema=schema, layer="layer1") as dst:
        dst.writerecords(features[0:5])
    with fiona.open(path, "w", driver="GPKG", schema=schema, layer="layer2") as dst:
        dst.writerecords(features[5:])

    with open(path, "rb") as f:
        with fiona.open(f, layer="layer1") as src:
            assert src.name == "layer1"
            assert len(src) == 5
    # Bug reported in #781 where this next section would fail
    with open(path, "rb") as f:
        with fiona.open(f, layer="layer2") as src:
            assert src.name == "layer2"
            assert len(src) == 62


def test_append_bytesio_exception(data_coutwildrnp_json):
    """Append is not supported, see #1027."""
    with pytest.raises(OSError):
        fiona.open(BytesIO(data_coutwildrnp_json), "a")


def test_mapinfo_raises():
    """Reported to be a crasher in #937"""
    driver = "MapInfo File"
    schema = {"geometry": "Point", "properties": OrderedDict([("position", "str")])}

    with BytesIO() as fout:
        with pytest.raises(OSError):
            with fiona.open(fout, "w", driver=driver, schema=schema) as collection:
                collection.write(
                    {
                        "type": "Feature",
                        "geometry": {"type": "Point", "coordinates": (0, 0)},
                        "properties": {"position": "x"},
                    }
                )


# TODO remove exclusion of MapInfo File once testdata_generator is fixed
@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in supported_drivers
        if _driver_supports_mode(driver, "w")
        and supports_vsi(driver)
        and driver not in {"MapInfo File"}
    ],
)
def test_write_memoryfile_drivers(driver, testdata_generator):
    """ Test if driver is able to write to memoryfile """
    range1 = list(range(0, 5))
    schema, crs, records1, _, _ = testdata_generator(driver, range1, [])

    with MemoryFile() as memfile:
        with memfile.open(driver=driver, schema=schema) as c:
            c.writerecords(records1)

        with memfile.open(driver=driver) as c:
            assert driver == c.driver
            items = list(c)
            assert len(items) == len(range1)


def test_multiple_layer_memoryfile(testdata_generator):
    """ Test ability to create multiple layers in memoryfile"""
    driver = "GPKG"
    range1 = list(range(0, 5))
    range2 = list(range(5, 10))
    schema, crs, records1, records2, _ = testdata_generator(driver, range1, range2)

    with MemoryFile() as memfile:
        with memfile.open(mode='w', driver=driver, schema=schema, layer="layer1") as c:
            c.writerecords(records1)
        with memfile.open(mode='w', driver=driver, schema=schema, layer="layer2") as c:
            c.writerecords(records2)

        with memfile.open(driver=driver, layer="layer1") as c:
            assert driver == c.driver
            items = list(c)
            assert len(items) == len(range1)

        with memfile.open(driver=driver, layer="layer2") as c:
            assert driver == c.driver
            items = list(c)
            assert len(items) == len(range1)


# TODO remove exclusion of MapInfo File once testdata_generator is fixed
@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in supported_drivers
        if _driver_supports_mode(driver, "a")
        and supports_vsi(driver)
        and driver not in {"MapInfo File"}
    ],
)
def test_append_memoryfile_drivers(driver, testdata_generator):
    """Test if driver is able to append to memoryfile"""
    range1 = list(range(0, 5))
    range2 = list(range(5, 10))
    schema, crs, records1, records2, _ = testdata_generator(driver, range1, range2)

    with MemoryFile() as memfile:
        with memfile.open(driver=driver, schema=schema) as c:
            c.writerecords(records1)

        with memfile.open(mode='a', driver=driver, schema=schema) as c:
            c.writerecords(records2)

        with memfile.open(driver=driver) as c:
            assert driver == c.driver
            items = list(c)
            assert len(items) == len(range1 + range2)


def test_memoryfile_driver_does_not_support_vsi():
    """An exception is raised with a driver that does not support VSI"""
    if "FileGDB" not in supported_drivers:
        pytest.skip("FileGDB driver not available")
    with pytest.raises(DriverError):
        with MemoryFile() as memfile:
            with memfile.open(driver="FileGDB"):
                pass


@pytest.mark.parametrize('mode', ['r', 'a'])
def test_modes_on_non_existing_memoryfile(mode):
    """Non existing memoryfile cannot opened in r or a mode"""
    with MemoryFile() as memfile:
        with pytest.raises(IOError):
            with memfile.open(mode=mode):
                pass


def test_write_mode_on_non_existing_memoryfile(profile_first_coutwildrnp_shp):
    """Exception is raised if a memoryfile is opened in write mode on a non empty memoryfile"""
    profile, first = profile_first_coutwildrnp_shp
    profile['driver'] = 'GeoJSON'
    with MemoryFile() as memfile:
        with memfile.open(**profile) as col:
            col.write(first)
        with pytest.raises(IOError):
            with memfile.open(mode="w"):
                pass


@requires_gpkg
def test_read_multilayer_memoryfile(path_coutwildrnp_json, tmpdir):
    """Test read access to multilayer dataset in from file-like object"""
    with fiona.open(path_coutwildrnp_json, "r") as src:
        schema = src.schema
        features = list(src)

    path = os.path.join(tmpdir, "test.gpkg")
    with fiona.open(path, "w", driver="GPKG", schema=schema, layer="layer1") as dst:
        dst.writerecords(features[0:5])
    with fiona.open(path, "w", driver="GPKG", schema=schema, layer="layer2") as dst:
        dst.writerecords(features[5:])

    with open(path, "rb") as f:
        with fiona.open(f, layer="layer1") as src:
            assert src.name == "layer1"
            assert len(src) == 5
    # Bug reported in #781 where this next section would fail
    with open(path, "rb") as f:
        with fiona.open(f, layer="layer2") as src:
            assert src.name == "layer2"
            assert len(src) == 62


def test_allow_unsupported_drivers(monkeypatch):
    """Test if allow unsupported drivers works as expected"""

    # We delete a known working driver from fiona.drvsupport so that we can use it
    monkeypatch.delitem(fiona.drvsupport.supported_drivers, "GPKG")

    schema = {"geometry": "Polygon", "properties": {}}

    # Test that indeed we can't create a file without allow_unsupported_drivers
    with pytest.raises(DriverError):
        with MemoryFile() as memfile:
            with memfile.open(mode="w", driver="GPKG", schema=schema):
                pass

    # Test that we can create file with allow_unsupported_drivers=True
    try:
        with MemoryFile() as memfile:
            with memfile.open(
                mode="w", driver="GPKG", schema=schema, allow_unsupported_drivers=True
            ):
                pass
    except Exception as e:
        assert (
            False
        ), f"Using allow_unsupported_drivers=True should not raise an exception: {e}"


def test_listdir_zipmemoryfile(bytes_coutwildrnp_zip):
    """Test list directories of a zipped memory file."""
    with ZipMemoryFile(bytes_coutwildrnp_zip) as memfile:
        assert memfile.listdir() == [
            "coutwildrnp.shp",
            "coutwildrnp.shx",
            "coutwildrnp.dbf",
            "coutwildrnp.prj",
        ]


def test_listlayers_zipmemoryfile(bytes_coutwildrnp_zip):
    """Test layers of a zipped memory file."""
    with ZipMemoryFile(bytes_coutwildrnp_zip) as memfile:
        assert memfile.listlayers() == ["coutwildrnp"]


def test_listdir_gdbzipmemoryfile(bytes_testopenfilegdb_zip):
    """Test list directories of a zipped GDB memory file."""
    with ZipMemoryFile(bytes_testopenfilegdb_zip, ext=".gdb.zip") as memfile:
        assert memfile.listdir() == [
            "testopenfilegdb.gdb",
        ]


def test_listdir_gdbzipmemoryfile_bis(bytes_testopenfilegdb_zip):
    """Test list directories of a zipped GDB memory file."""
    with ZipMemoryFile(bytes_testopenfilegdb_zip, filename="temp.gdb.zip") as memfile:
        assert memfile.listdir() == [
            "testopenfilegdb.gdb",
        ]



"""Tests for `$ fio filter`."""

from fiona.fio.main import main_group


def test_fail(runner):
    result = runner.invoke(main_group, ['filter',
        "f.properties.test > 5"
    ], "{'type': 'no_properties'}")
    assert result.exit_code == 1


def test_seq(feature_seq, runner):

    result = runner.invoke(main_group, ['filter',
        "f.properties.AREA > 0.01"], feature_seq, catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output.count('Feature') == 2

    result = runner.invoke(main_group, ['filter',
        "f.properties.AREA > 0.015"], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('Feature') == 1

    result = runner.invoke(main_group, ['filter',
        "f.properties.AREA > 0.02"], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('Feature') == 0



"""OGR 64bit handling: https://trac.osgeo.org/gdal/wiki/rfc31_ogr_64

Shapefile: OFTInteger fields are created by default with a width of 9
characters, so to be unambiguously read as OFTInteger (and if specifying
integer that require 10 or 11 characters. the field is dynamically extended
like managed since a few versions). OFTInteger64 fields are created by default
with a width of 18 digits, so to be unambiguously read as OFTInteger64, and
extented to 19 or 20 if needed. Integer fields of width between 10 and 18
will be read as OFTInteger64. Above they will be treated as OFTReal. In
previous GDAL versions, Integer fields were created with a default with of 10,
and thus will be now read as OFTInteger64. An open option, DETECT_TYPE=YES, can
be specified so as OGR does a full scan of the DBF file to see if integer
fields of size 10 or 11 hold 32 bit or 64 bit values and adjust the type
accordingly (and same for integer fields of size 19 or 20, in case of overflow
of 64 bit integer, OFTReal is chosen)
"""

import pytest

import fiona
from fiona.env import calc_gdal_version_num, get_gdal_version_num
from fiona.model import Feature


def testCreateBigIntSchema(tmpdir):
    name = str(tmpdir.join("output1.shp"))

    a_bigint = 10 ** 18 - 1
    fieldname = "abigint"

    kwargs = {
        "driver": "ESRI Shapefile",
        "crs": "EPSG:4326",
        "schema": {"geometry": "Point", "properties": [(fieldname, "int:10")]},
    }

    with fiona.open(name, "w", **kwargs) as dst:
        rec = {}
        rec["geometry"] = {"type": "Point", "coordinates": (0, 0)}
        rec["properties"] = {fieldname: a_bigint}
        dst.write(Feature.from_dict(**rec))

    with fiona.open(name) as src:
        if fiona.gdal_version >= (2, 0, 0):
            first = next(iter(src))
            assert first["properties"][fieldname] == a_bigint


@pytest.mark.parametrize("dtype", ["int", "int64"])
def test_issue691(tmpdir, dtype):
    """Type 'int' maps to 'int64'"""
    schema = {"geometry": "Any", "properties": {"foo": dtype}}
    with fiona.open(
        str(tmpdir.join("test.shp")),
        "w",
        driver="Shapefile",
        schema=schema,
        crs="epsg:4326",
    ) as dst:
        dst.write(
            Feature.from_dict(
                **{
                    "type": "Feature",
                    "geometry": {
                        "type": "Point",
                        "coordinates": (-122.278015, 37.868995),
                    },
                    "properties": {"foo": 3694063472},
                }
            )
        )

    with fiona.open(str(tmpdir.join("test.shp"))) as src:
        assert src.schema["properties"]["foo"] == "int:18"
        first = next(iter(src))
        assert first["properties"]["foo"] == 3694063472



import platform
import sys

import fiona
from fiona._env import get_gdal_release_name, get_proj_version_tuple


def show_versions():
    """
    Prints information useful for bug reports
    """

    fiona_version = fiona.__version__
    gdal_release_name = get_gdal_release_name()
    proj_version_tuple = get_proj_version_tuple()
    proj_version = ".".join(map(str, proj_version_tuple))
    os_info = "{system} {release}".format(system=platform.system(),
                                          release=platform.release())
    python_version = platform.python_version()
    python_exec = sys.executable

    msg = ("Fiona version: {fiona_version}"
           "\nGDAL version: {gdal_release_name}"
           "\nPROJ version: {proj_version}"
           "\n"
           "\nOS: {os_info}"
           "\nPython: {python_version}"
           "\nPython executable: '{python_exec}'"
           "\n"
           )

    print(msg.format(fiona_version=fiona_version,
                     gdal_release_name=gdal_release_name,
                     proj_version=proj_version,
                     os_info=os_info,
                     python_version=python_version,
                     python_exec=python_exec))



"""Logging helper classes."""

import logging


class FieldSkipLogFilter(logging.Filter):
    """Filter field skip log messges.

    At most, one message per field skipped per loop will be passed.
    """

    def __init__(self, name=''):
        super().__init__(name)
        self.seen_msgs = set()

    def filter(self, record):
        """Pass record if not seen."""
        msg = record.getMessage()
        if msg.startswith("Skipping field"):
            retval = msg not in self.seen_msgs
            self.seen_msgs.add(msg)
            return retval
        else:
            return 1


class LogFiltering:
    def __init__(self, logger, filter):
        self.logger = logger
        self.filter = filter

    def __enter__(self):
        self.logger.addFilter(self.filter)

    def __exit__(self, *args, **kwargs):
        self.logger.removeFilter(self.filter)



import json

import click
from cligj import use_rs_opt

from .helpers import obj_gen, eval_feature_expression
from fiona.fio import with_context_env
from fiona.model import ObjectEncoder


@click.command(short_help="Calculate GeoJSON property by Python expression")
@click.argument('property_name')
@click.argument('expression')
@click.option('--overwrite', is_flag=True, default=False,
              help="Overwrite properties, default: False")
@use_rs_opt
@click.pass_context
@with_context_env
def calc(ctx, property_name, expression, overwrite, use_rs):
    """
    Create a new property on GeoJSON features using the specified expression.

    \b
    The expression is evaluated in a restricted namespace containing:
        - sum, pow, min, max and the imported math module
        - shape (optional, imported from shapely.geometry if available)
        - bool, int, str, len, float type conversions
        - f (the feature to be evaluated,
             allows item access via javascript-style dot notation using munch)

    The expression will be evaluated for each feature and its
    return value will be added to the properties
    as the specified property_name. Existing properties will not
    be overwritten by default (an Exception is raised).

    Example

    \b
    $ fio cat data.shp | fio calc sumAB  "f.properties.A + f.properties.B"

    """
    stdin = click.get_text_stream('stdin')
    source = obj_gen(stdin)

    for i, obj in enumerate(source):
        features = obj.get("features") or [obj]

        for j, feat in enumerate(features):

            if not overwrite and property_name in feat["properties"]:
                raise click.UsageError(
                    "{} already exists in properties; "
                    "rename or use --overwrite".format(property_name)
                )

            feat["properties"][property_name] = eval_feature_expression(
                feat, expression
            )

            if use_rs:
                click.echo("\x1e", nl=False)

            click.echo(json.dumps(feat, cls=ObjectEncoder))



import logging
import os
import shutil
import sys
import tempfile
from collections import OrderedDict
import unittest

import pytest

import fiona
from fiona.errors import SchemaError
from fiona.model import Feature


class TestUnicodePath(unittest.TestCase):
    def setUp(self):
        tempdir = tempfile.mkdtemp()
        self.dir = os.path.join(tempdir, "français")
        shutil.copytree(os.path.join(os.path.dirname(__file__), "data"), self.dir)

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.dir))

    def test_unicode_path(self):
        path = self.dir + "/coutwildrnp.shp"
        with fiona.open(path) as c:
            assert len(c) == 67

    def test_unicode_path_layer(self):
        path = self.dir
        layer = "coutwildrnp"
        with fiona.open(path, layer=layer) as c:
            assert len(c) == 67

    def test_utf8_path(self):
        path = self.dir + "/coutwildrnp.shp"
        if sys.version_info < (3,):
            with fiona.open(path) as c:
                assert len(c) == 67


class TestUnicodeStringField(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @pytest.mark.xfail(reason="OGR silently fails to convert strings")
    def test_write_mismatch(self):
        """TOFIX: OGR silently fails to convert strings"""
        # Details:
        #
        # If we tell OGR that we want a latin-1 encoded output file and
        # give it a feature with a unicode property that can't be converted
        # to latin-1, no error is raised and OGR just writes the utf-8
        # encoded bytes to the output file.
        #
        # This might be shapefile specific.
        #
        # Consequences: no error on write, but there will be an error
        # on reading the data and expecting latin-1.
        schema = {"geometry": "Point", "properties": {"label": "str", "num": "int"}}

        with fiona.open(
            os.path.join(self.tempdir, "test-write-fail.shp"),
            "w",
            driver="ESRI Shapefile",
            schema=schema,
            encoding="latin1",
        ) as c:
            c.writerecords(
                [
                    {
                        "type": "Feature",
                        "geometry": {"type": "Point", "coordinates": [0, 0]},
                        "properties": {"label": "徐汇区", "num": 0},
                    }
                ]
            )

        with fiona.open(os.path.join(self.tempdir), encoding="latin1") as c:
            f = next(iter(c))
            # Next assert fails.
            assert f.properties["label"] == "徐汇区"

    def test_write_utf8(self):
        schema = {
            "geometry": "Point",
            "properties": {"label": "str", "verit\xe9": "int"},
        }
        with fiona.open(
            os.path.join(self.tempdir, "test-write.shp"),
            "w",
            "ESRI Shapefile",
            schema=schema,
            encoding="utf-8",
        ) as c:
            c.writerecords(
                [
                    Feature.from_dict(
                        **{
                            "type": "Feature",
                            "geometry": {"type": "Point", "coordinates": [0, 0]},
                            "properties": {"label": "Ba\u2019kelalan", "verit\xe9": 0},
                        }
                    )
                ]
            )

        with fiona.open(os.path.join(self.tempdir), encoding="utf-8") as c:
            f = next(iter(c))
            assert f.properties["label"] == "Ba\u2019kelalan"
            assert f.properties["verit\xe9"] == 0

    @pytest.mark.iconv
    def test_write_gb18030(self):
        """Can write a simplified Chinese shapefile"""
        schema = {"geometry": "Point", "properties": {"label": "str", "num": "int"}}
        with fiona.open(
            os.path.join(self.tempdir, "test-write-gb18030.shp"),
            "w",
            driver="ESRI Shapefile",
            schema=schema,
            encoding="gb18030",
        ) as c:
            c.writerecords(
                [
                    Feature.from_dict(
                        **{
                            "type": "Feature",
                            "geometry": {"type": "Point", "coordinates": [0, 0]},
                            "properties": {"label": "徐汇区", "num": 0},
                        }
                    )
                ]
            )

        with fiona.open(os.path.join(self.tempdir), encoding="gb18030") as c:
            f = next(iter(c))
            assert f.properties["label"] == "徐汇区"
            assert f.properties["num"] == 0

    @pytest.mark.iconv
    def test_gb2312_field_wrong_encoding(self):
        """Attempt to create field with a name not supported by the encoding

        ESRI Shapefile driver defaults to ISO-8859-1 encoding if none is
        specified. This doesn't support the field name used. Previously this
        went undetected and would raise a KeyError later when the user tried
        to write a feature to the layer. Instead we raise a more useful error.

        See GH#595.
        """
        field_name = "区县名称"
        meta = {
            "schema": {
                "properties": OrderedDict([(field_name, "int")]),
                "geometry": "Point",
            },
            "driver": "ESRI Shapefile",
        }
        feature = Feature.from_dict(
            **{
                "properties": {field_name: 123},
                "geometry": {"type": "Point", "coordinates": [1, 2]},
            }
        )
        # when encoding is specified, write is successful
        with fiona.open(
            os.path.join(self.tempdir, "test1.shp"), "w", encoding="GB2312", **meta
        ) as collection:
            collection.write(feature)
        # no encoding
        with pytest.raises(SchemaError):
            fiona.open(os.path.join(self.tempdir, "test2.shp"), "w", **meta)



"""Unittests to verify Fiona is functioning properly with other software."""


from collections import UserDict

import fiona
from fiona.model import Feature


def test_dict_subclass(tmpdir):
    """Rasterio now has a `CRS()` class that subclasses
    `collections.UserDict()`.  Make sure we can receive it.

    `UserDict()` is a good class to test against because in Python 2 it is
    not a subclass of `collections.Mapping()`, so it provides an edge case.
    """

    class CRS(UserDict):
        pass

    outfile = str(tmpdir.join("test_UserDict.geojson"))

    profile = {
        "crs": CRS(init="EPSG:4326"),
        "driver": "GeoJSON",
        "schema": {"geometry": "Point", "properties": {}},
    }

    with fiona.open(outfile, "w", **profile) as dst:
        dst.write(
            Feature.from_dict(
                **{
                    "type": "Feature",
                    "properties": {},
                    "geometry": {"type": "Point", "coordinates": (10, -10)},
                }
            )
        )

    with fiona.open(outfile) as src:
        assert len(src) == 1
        assert src.crs == {"init": "epsg:4326"}



import pytest

import fiona
from .test_collection import TestReading


def test_index_selection(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp, 'r', layer=0) as c:
        assert len(c) == 67


class TestFileReading(TestReading):
    @pytest.fixture(autouse=True)
    def shapefile(self, path_coutwildrnp_shp):
        self.c = fiona.open(path_coutwildrnp_shp, 'r', layer='coutwildrnp')
        yield
        self.c.close()

    def test_open_repr(self, path_coutwildrnp_shp):
        assert (
            repr(self.c) ==
            ("<open Collection '{path}:coutwildrnp', mode 'r' "
             "at {id}>".format(path=path_coutwildrnp_shp, id=hex(id(self.c)))))

    def test_closed_repr(self, path_coutwildrnp_shp):
        self.c.close()
        assert (
            repr(self.c) ==
            ("<closed Collection '{path}:coutwildrnp', mode 'r' "
             "at {id}>".format(path=path_coutwildrnp_shp, id=hex(id(self.c)))))

    def test_name(self):
        assert self.c.name == 'coutwildrnp'


class TestDirReading(TestReading):
    @pytest.fixture(autouse=True)
    def shapefile(self, data_dir):
        self.c = fiona.open(data_dir, "r", layer="coutwildrnp")
        yield
        self.c.close()

    def test_open_repr(self, data_dir):
        assert (
            repr(self.c) ==
            ("<open Collection '{path}:coutwildrnp', mode 'r' "
             "at {id}>".format(path=data_dir, id=hex(id(self.c)))))

    def test_closed_repr(self, data_dir):
        self.c.close()
        assert (
            repr(self.c) ==
            ("<closed Collection '{path}:coutwildrnp', mode 'r' "
             "at {id}>".format(path=data_dir, id=hex(id(self.c)))))

    def test_name(self):
        assert self.c.name == 'coutwildrnp'

    def test_path(self, data_dir):
        assert self.c.path == data_dir


def test_invalid_layer(path_coutwildrnp_shp):
    with pytest.raises(ValueError):
        fiona.open(path_coutwildrnp_shp, layer="foo")


def test_write_invalid_numeric_layer(path_coutwildrnp_shp, tmpdir):
    with pytest.raises(ValueError):
        fiona.open(str(tmpdir.join("test-no-iter.shp")), mode='w', layer=0)



"""Unittests for $ fio dump"""


import json

from click.testing import CliRunner
import pytest

import fiona
from fiona.fio.main import main_group


def test_dump(path_coutwildrnp_shp):
    runner = CliRunner()
    result = runner.invoke(main_group, ['dump', path_coutwildrnp_shp])
    assert result.exit_code == 0
    assert '"FeatureCollection"' in result.output


@pytest.mark.parametrize("layer", ["routes", "1", "tracks", "track_points"])
def test_dump_layer(path_gpx, layer):
    runner = CliRunner()
    result = runner.invoke(main_group, ["dump", path_gpx, "--layer", layer])
    assert result.exit_code == 0
    assert '"FeatureCollection"' in result.output


def test_dump_layer_vfs(path_coutwildrnp_zip):
    path = f"zip://{path_coutwildrnp_zip}"
    result = CliRunner().invoke(main_group, ["dump", path])
    assert result.exit_code == 0
    loaded = json.loads(result.output)
    with fiona.open(path) as src:
        assert len(loaded['features']) == len(src)
        assert len(loaded['features']) > 0
